"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import os
import socket

from sailthru import SailthruClient

from sdc_dp_helpers.api_utilities.date_managers import date_range, date_handler
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.sailthru.config_magagers import get_config


class CustomSailThruReader:
    """
    Custom SailThru Reader
    """

    def __init__(self, config_filepath, creds_filepath):
        self.config = get_config(config_filepath)
        self.credentials = get_config(creds_filepath)

        self.sailthru_client = SailthruClient(
            api_key=self.credentials.get("api_key", None),
            secret=self.credentials.get("api_secret", None),
        )

    def _get_stat_templates(self):
        """
        Dynamically gathers all templates for 'send' stat.
        """
        response = self.sailthru_client.api_get(
            "template",
            {
                "start_date": date_handler(self.config["send"]["start_date"]),
                "end_date": date_handler(self.config["send"]["end_date"]),
            },
        )
        if response.is_ok():
            template_names = []
            template_data = response.get_body()
            for templates in template_data.get("templates"):
                template_names.append(templates.get("name"))

            return template_names

        raise EnvironmentError(
            f"Error: {response.get_error().get_error_code()}\n"
            f"Mesage: {response.get_error().get_message()}"
        )

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def _get_stat(self, data, action="stats"):
        response = self.sailthru_client.api_get(action=action, data=data)
        if response.is_ok():
            return response.get_body()
        elif not response.is_ok():
            error = response.get_error()
            message = "APIError:{}, Status Code:{}, Error Code:{}".format(
                error.get_message(),
                str(response.get_status_code()),
                str(error.get_error_code()),
            )

            if (
                str(response.get_status_code()) != "404"
                and str(error.get_error_code()) != "99"
            ):
                raise EnvironmentError(message)

            # If there is no data for something like a template,
            # just a warning is printed.
            print(f"Warning: No data in given instance.\n{message}.")
        else:
            return None

    @retry_handler(
        exceptions=(socket.timeout), total_tries=5, initial_wait=60, backoff_factor=5
    )
    def run_query(self):
        """
        Request various stats from Sailthru about primary list membership
        or campaign and triggered message activity.
        Endpoint URL: https://api.sailthru.com/stats
        Additional parameters are dependent on the stat value
        the type of stats you want to request:
            - list
            - blast
            - send
        """
        _data_set = []
        for action, data in self.config.items():
            _date_range_list = date_range(
                start_date=data["start_date"], end_date=data["end_date"]
            )

            if action == "blast":
                """
                Retrieve information about a particular campaign or aggregated information
                from all campaigns over a specified date range.
                """
                for _date in _date_range_list:
                    print(f"Gathering BLAST data for date: {_date}.")
                    stat_data = self._get_stat(
                        data={
                            "stat": "blast",
                            "start_date": _date,
                            "end_date": _date,
                        }
                    )

                    if stat_data is not None:
                        stat_data["date"] = _date
                        stat_data["stat"] = "blast"
                        _data_set.append(stat_data)

            elif action == "list":
                """
                Retrieve information about your subscriber counts on all lists or a particular
                list, as counted on a specified day’s snapshot, defaulting to the current day.
                """
                for _date in _date_range_list:
                    print(f"Gathering LIST data for date: {_date}.")
                    stat_data = self._get_stat(
                        data={
                            "stat": "list",
                            "date": _date,
                        }
                    )

                    if stat_data is not None:
                        stat_data["stat"] = "list"
                        _data_set.append(stat_data)

            elif action == "send":
                """
                Retrieve information about a particular triggered message or aggregated information
                from triggered messages from that template over a specified date range.
                """
                _template_names = self._get_stat_templates()
                print(f"Total templates: {len(_template_names)}.")
                for _date in _date_range_list:
                    for template_name in _template_names:
                        print(
                            f"Gathering SEND data at template: {template_name}. For date: {_date}."
                        )
                        stat_data = self._get_stat(
                            data={
                                "stat": "send",
                                "template": template_name,
                                "start_date": _date,
                                "end_date": _date,
                            }
                        )

                        if stat_data is not None:
                            stat_data["stat"] = "send"
                            stat_data["template"] = template_name
                            _data_set.append(stat_data)

            else:
                raise EnvironmentError(
                    f"Invalid action: {action} for stat endpoint. "
                    "Please use blast, list or send."
                )

        return _data_set



# pylint: disable=too-few-public-methods

"""
    CUSTOM WRITER CLASSES
"""
import json
from datetime import datetime

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, profile_name=None):
        self.bucket = bucket

        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource("s3")

    def write_to_s3(self, json_data, config, partition, layout="standard"):
        """
        Construct partitioning and file name conventions in s3
        according to business specifications, and write to S3.
        """
        module = config.get("module").lower()
        partial_pull = config.get("partial_pull", True)
        organisation_name = config.get("organization_name").lower()
        partial_pull_size = config.get("partial_pull_size", 5)
        now = datetime.now().strftime("%Y%m%d%H%M")

        if partial_pull:
            key_path = f"{organisation_name}/last_{partial_pull_size}_pages/{module}/{module}_{partition}_{layout}_{now}.json"
        else:
            key_path = f"{organisation_name}/full/{module}/{module}_{partition}_{layout}_full.json"

        print(f"Write path: S3://{self.bucket}/{key_path}.")

        self.s3_resource.Object(self.bucket, key_path).put(Body=json.dumps(json_data))



"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import gzip
import io
import json
import os

import boto3
from interface import implements

from sdc_dp_helpers.base_writer import BaseWriter, CustomLocalJsonWriter


class GAV3Writer:
    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.folder_path = folder_path

    @staticmethod
    def _partition_by_date(dataset, date_key):
        partition_dataset = {}
        for row in dataset:
            date_value = row.get(date_key)

            if date_value not in partition_dataset:
                partition_dataset[date_value] = []
            partition_dataset[date_value].append(row)

        return partition_dataset

    @staticmethod
    def _partition_by_profile(dataset):
        partition_dataset = {}
        for row in dataset:
            date_value = row.get("profileId")

            if date_value not in partition_dataset:
                partition_dataset[date_value] = []
            partition_dataset[date_value].append(row)

        return partition_dataset

    def write_partitions_to_s3(self, dataset, date_key="ga:date"):
        """
        If we pull data over a period of time for the same dimensions, we will
        begin to see discrepancies in the data.

        This is because the sampling means that if we pull one set of data on different
        days, the metrics might be slightly different, and we end up with duplication.

        So the plan is to ensure that if we:
            - pull from a specific date, we overwrite all that days data in the buckets.
            - this way, we ensure all data is original and in its most updated state.
        """
        data_by_view_id = self._partition_by_profile(dataset)
        for view_id, view_id_dataset in data_by_view_id.items():
            data_by_date = self._partition_by_date(view_id_dataset, date_key=date_key)
            for date, date_dataset in data_by_date.items():
                write_path = f"{self.folder_path}/{view_id}/{date}.json"
                print(
                    f"Writing data to s3://{write_path} partitioned by view_id and date."
                )
                self.s3_resource.Object(self.bucket, write_path).put(
                    Body=json.dumps(date_dataset)
                )


class CustomS3GZJsonWriter(CustomLocalJsonWriter, implements(BaseWriter)):
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(
        self,
        file_name: str,
        folder_path: str,
        bucket: str,
        profile_name: str = None,
        **kwargs,
    ):
        self.os_path_sep = "/"

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.bucket = bucket
        """Writes a general object to s3"""
        self.s3_resource = self.boto3_session.resource("s3")
        super().__init__(file_name=file_name, folder_path=folder_path, **kwargs)
        self.metadata_file = kwargs.get("metadata_file", "metadata.json.gz")

    def set_full_path(self, nested_dirs: list = None):
        """Set full path to write to"""
        # creates full path
        full_path = [self.folder_path] + nested_dirs
        return os.path.join(*full_path).replace("\\", "/")

    # pylint: disable=no-member,too-many-arguments
    def write_to_s3(
        self,
        json_data,
        data_path,
        mode="w",
        encoding="utf-8",
        ensure_ascii=False,
        default=None,
    ):
        """
        Write report to s3 bucket
        """
        buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode=mode) as file_:
            with io.TextIOWrapper(file_, encoding=encoding) as textio_obj:
                textio_obj.write(
                    json.dumps(json_data, ensure_ascii=ensure_ascii, default=default)
                )
        # write to s3
        buffer.seek(0)
        self.s3_resource.Bucket(self.bucket).put_object(
            Key=os.path.join(self.full_path, data_path).replace("\\", "/"), Body=buffer
        )

    def write_metadata(self, metadata: dict):
        """ "
        Creates required GA/ViewID/metricName structure
        """
        # create directory structure if not exists
        self.full_path = self.set_full_path(
            nested_dirs=[metadata.get("viewID", "."), metadata.get("metricName", "")]
        )
        self.write_to_s3(json_data=metadata, data_path=self.metadata_file, mode="w")

    def write_file(self):
        """
        upload python dict into s3 bucket with gzip archive
        """
        # generate a unix timestamp suffix for file
        suff = self.timestamp_suffix()
        data_path = "{}_{}.json.gz".format(self.file_name, suff)
        self.write_to_s3(self.data, data_path, mode="wb")
        self.data = []



import json

import yaml


def get_config(configs, fmt):
    """
    Gathers query configurations from json or yaml.
    """
    if configs is not None:
        if fmt in ["yaml", "yml"]:
            with open(configs, "r") as file:
                config = yaml.safe_load(file)
        elif fmt == "json":
            with open(configs, "r") as file:
                config = json.load(file)

    return config
